/*
firestring.c - FireString replacement string functions
Copyright (C) 2000 Ian Gulliver

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Parts of this library are taken from dstr and dconf, part of LibD:
* Copyright (C) 2001-2002  Andres Salomon <dilinger@mp3revolution.net>

*/

#include "firestring.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>

static const char tagstring[] = "$Id: firestring.c,v 1.57 2002/06/24 19:58:16 ian Exp $";

void *firestring_malloc(const size_t size) {
	char *output;
	output = malloc(size);
	if (output == NULL) {
		perror("malloc");
		exit(EXIT_FAILURE);
	}
	return output;
}

void *firestring_realloc(void *old, const size_t new) {
	char *output;
	output = realloc(old,new);
	if (output == NULL) {
		perror("realloc");
		exit(EXIT_FAILURE);
	}
	return output;
}

char *firestring_strdup(const char * const input) {
	char *output;
	size_t s;
	if (input == NULL)
		return NULL;
	s = strlen(input) + 1;
	output = firestring_malloc(s);
	memcpy(output,input,s);
	return(output);
}

void firestring_strncpy(char * const to, const char * const from, const size_t size) {
	strncpy(to,from,size);
	to[size - 1]= '\0';
	return;
}

void firestring_strncat(char * const to, const char * const from, const size_t size) {
	size_t l;
	l = strlen(to);
	firestring_strncpy(&to[l],from,size - l);
	return;
}

static int shownum_unsigned(unsigned long long m, int padzero, int numpad, char *numbuf, int space) {
	int power; /* smallest power of 10 greater than m */
	long long multiple = 1;
	int r;
	char f;
	int i = 0;

	for (power = 0; multiple <= m; power++)
		multiple *= 10;

	if (m == 0) {
		padzero = 1;
		numpad = 1;
	}

	f = padzero == 1 ? '0' : ' ';

	if (power > space || numpad > space)
		return -1;

	if (power < numpad) {
		numpad -= power;
		for (; i < numpad; i++)
			numbuf[i] = f;
	}

	for (;power > 0; power--) {
		multiple /= 10;
		r = m / multiple;
		numbuf[i++] = '0' + r;
		m -= r * multiple;
	}
	return i;
}

static int shownum_signed(long long n, int padzero, int numpad, char *numbuf, int space) {
	unsigned long long m;
	int i = 0;

#define ABS(n) ((n < 0) ? n * -1 : n);
	m = (unsigned long long) ABS(n);
	if ((long long) m != n) {
		numbuf[0] = '-';
		i = shownum_unsigned(m,padzero,numpad,&numbuf[1],space - 1);
		if (i == -1)
			return -1;
		else
			return i;
	} else
		return shownum_unsigned(m,padzero,numpad,numbuf,space);
}

int firestring_snprintf(char * out, const size_t size, const char * const format, ...) {
	va_list ap;
	size_t f,o = 0,fl,tl,ml;
	char *tempchr;
	struct firestring_estr_t *e;
	int b = 0;
	int padzero, numpad;
	int m;

	fl = strlen(format);
	ml = size - 1;

#define SIGNED_NUMBER(type) \
	m = shownum_signed((long long) va_arg(ap,type),padzero,numpad,&out[o],size - o - 1); \
	if (m == -1) \
		b = 1; \
	else \
		o += m;

#define UNSIGNED_NUMBER(type) \
	m = shownum_unsigned((unsigned long long) va_arg(ap,type),padzero,numpad,&out[o],size - o - 1); \
	if (m == -1) \
		b = 1; \
	else \
		o += m;

	va_start(ap,format);
	for (f = 0; f < fl && o < ml && b == 0; f++) {
		if (format[f] == '%') {
			padzero = 0;
			numpad = 0;
formatcheck:
			switch(format[++f]) {
				case '0':
					if (padzero == 0) {
						padzero = 1;
						goto formatcheck;
					}
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					numpad = (numpad * 10) + (format[f] - '0');
					goto formatcheck;
				case 's':
					tempchr = va_arg(ap,char *);
					if (tempchr == NULL)
						tempchr = "(null)";
					tl = strlen(tempchr);
					if (tl + o > ml)
						b = 1;
					else {
						memcpy(&out[o],tempchr,tl);
						o += tl;
					}
					break;
				case 'e':
					e = va_arg(ap,struct firestring_estr_t *);
					if (e == NULL)
						break;
					if (e->l + o > ml)
						b = 1;
					else {
						memcpy(&out[o],e->s,e->l);
						o += e->l;
					}
					break;
				case 'd': /* signed int */
					SIGNED_NUMBER(int)
					break;
				case 'l': /* signed long */
					SIGNED_NUMBER(long)
					break;
				case 'u': /* unsigned int */
					UNSIGNED_NUMBER(unsigned int)
					break;
				case 'y': /* unsigned long */
					UNSIGNED_NUMBER(unsigned long)
					break;
				case 'g': /* signed long long */
					SIGNED_NUMBER(long long)
					break;
				case 'o': /* unsigned long long */
					UNSIGNED_NUMBER(unsigned long long)
					break;
				case '%':
					out[o++] = '%';
					break;
			}
		} else
			out[o++] = format[f];
	}
	out[o] = '\0';
	return o;
}

int firestring_strncasecmp(const char * const s1, const char * const s2, const size_t n) {
	size_t s;
	for (s = 0; s < n; s++) {
		if (tolower((unsigned char) s1[s]) != tolower((unsigned char) s2[s])) {
			if (tolower((unsigned char) s1[s]) < tolower((unsigned char) s2[s]))
				return -1;
			else
				return 1;
		}
		if (s1[s] == '\0')
			return 0;
	}
	return 0;
}

int firestring_strcasecmp(const char * const s1, const char * const s2) {
	size_t s;
	char c1, c2;
	for (s = 0; (s1[s] != '\0') && (s2[s] != '\0') && (tolower((unsigned char) s1[s]) == tolower((unsigned char) s2[s])); s++);

	c1 = tolower((unsigned char) s1[s]);
	c2 = tolower((unsigned char) s2[s]);
	if (c1 == c2)
		return 0;
	else if (c1 < c2)
		return -1;
	else
		return 1;
}

char *firestring_concat (const char *s, ...) {
	const char *curr;
	size_t len = 0;
	va_list va;
	char *ret = NULL;

	/* get length */
	va_start(va,s);
	curr = s;
	while (curr != NULL) {
		len += strlen(curr);
		curr = va_arg(va, const char *);
	}
	va_end(va);

	/* allocate string */
	if (len) {
		ret = firestring_malloc(len + 1);
		*ret = '\0';
	}

	/* copy */
	va_start(va,s);
	curr = s;
	while (curr != NULL) {
		strcat(ret,curr);
		curr = va_arg(va, const char *);
	}
	va_end(va);

	return ret;
}

char *firestring_chomp(char *s) {
	char *ret = s;
	if (s != NULL) {
		s = &s[strlen(s) - 1];
		while (isspace(*s)) {
			*s = '\0';
			if (s == ret)
				break;
			s--;
		}
	}
	return ret;
}

char *firestring_chug(char *s) {
	if (s != NULL) {
		while (isspace(*s))
			s++;
	}
	return s;
}

void firestring_estr_alloc(struct firestring_estr_t *f, int a) {
	f->s = firestring_malloc(a + 1);
	f->a = a;
	f->l = 0;
}

void firestring_estr_free(struct firestring_estr_t *f) {
	free(f->s);
	f->a = 0;
	f->l = 0;
}

void firestring_estr_0(struct firestring_estr_t *f) {
	f->s[f->l] = '\0';
}

int firestring_estr_read(struct firestring_estr_t *f, int fd) {
	int i;
	if (f->l == f->a)
		return 2;
	i = read(fd,&f->s[f->l],f->a - f->l);
	if (i == -1 && errno == EAGAIN)
		return 0;
	if (i <= 0)
		return 1;
	f->l += i;
	return 0;
}

int firestring_estr_sprintf(struct firestring_estr_t *o, const char * const format, ...) {
	va_list ap;
	size_t f,fl,tl;
	char *tempchr;
	struct firestring_estr_t *e;
	int b = 0;
	int padzero, numpad;
	int m;

#undef SIGNED_NUMBER
#undef UNSIGNED_NUMBER

#define SIGNED_NUMBER(type) \
	m = shownum_signed((long long) va_arg(ap,type),padzero,numpad,&o->s[o->l],o->a - o->l); \
	if (m == -1) \
		b = 1; \
	else \
		o->l += m;

#define UNSIGNED_NUMBER(type) \
	m = shownum_unsigned((unsigned long long) va_arg(ap,type),padzero,numpad,&o->s[o->l],o->a - o->l); \
	if (m == -1) \
		b = 1; \
	else \
		o->l += m;

	fl = strlen(format);
	o->l = 0;

	va_start(ap,format);
	for (f = 0; f < fl && o->l < o->a && b == 0; f++) {
		if (format[f] == '%') {
			padzero = 0;
			numpad = 0;
eformatcheck:
			switch(format[++f]) {
				case '0':
					if (padzero == 0) {
						padzero = 1;
						goto eformatcheck;
					}
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					numpad = (numpad * 10) + (format[f-1] - '0');
					goto eformatcheck;
				case 's':
					tempchr = va_arg(ap,char *);
					if (tempchr == NULL)
						tempchr = "(null)";
					tl = strlen(tempchr);
					if (tl + o->l > o->a)
						b = 1;
					else {
						memcpy(&o->s[o->l],tempchr,tl);
						o->l += tl;
					}
					break;
				case 'e':
					e = va_arg(ap,struct firestring_estr_t *);
					if (e == NULL)
						break;
					if (e->l + o->l > o->a)
						b = 1;
					else {
						memcpy(&o->s[o->l],e->s,e->l);
						o->l += e->l;
					}
					break;
				case 'd': /* signed int */
					SIGNED_NUMBER(int)
					break;
				case 'l': /* signed long */
					SIGNED_NUMBER(long)
					break;
				case 'u': /* unsigned int */
					UNSIGNED_NUMBER(unsigned int)
					break;
				case 'y': /* unsigned long */
					UNSIGNED_NUMBER(unsigned long)
					break;
				case 'g': /* signed long long */
					SIGNED_NUMBER(long long)
					break;
				case 'o': /* unsigned long long */
					UNSIGNED_NUMBER(unsigned long long)
					break;
				case '%':
					o->s[o->l++] = '%';
					break;
			}
		} else
			o->s[o->l++] = format[f];
	}
	return o->l;
}

static char *try_escaped_newline(char *s) {
	char *esc, *end;
	
	/* check for terminating '\\' */
	esc = end = strrchr (s, '\\');
	if (end) {
		end++;

		/* check if the only thing separating the '\\' and the end
		 * of the string is just space */
		while (*end && isspace (*end))
			end++;
		if (*end == '\0')
			*esc = '\0';
	}

	return s;
}

int firestring_estr_strchr(struct firestring_estr_t *f, const char c, const int start) {
	int i;

	for (i = start; i < f->l; i++)
		if (f->s[i] == c)
			return i;

	return -1;
}

int firestring_estr_strstr(struct firestring_estr_t *f, const char *s, const int start) {
	int i;
	int l;
	
	l = strlen(s);
	
	for (i = start; i <= f->l - l; i++)
		if (memcmp(&f->s[i],s,l) == 0)
			return i;

	return -1;
}

int firestring_estr_stristr(struct firestring_estr_t *f, const char *s, const int start) {
	int i;
	int j;
	int l;
	
	l = strlen(s);
	
	for (i = start; i <= f->l - l; i++) {
		for (j = 0; j < l; j++)
			if (tolower(f->s[i+j]) != tolower(s[j]))
					break;
		if (j == l)
			return i;
	}

	return -1;
}

int firestring_estr_starts(struct firestring_estr_t *f, const char *s) {
	int l;
	int i;

	l = strlen(s);
	
	if (f->l < l)
		return 1;

	for (i = 0; i < l; i++)
		if (tolower(f->s[i]) != tolower(s[i]))
			return 1;

	return 0;
}

int firestring_estr_ends(struct firestring_estr_t *f, const char *s) {
	int l;
	int i;
	int t;

	l = strlen(s);

	if (f->l < l)
		return 1;

	t = f->l - l;
	for (i = f->l - 1; i >= t; i--)
		if (tolower(f->s[i]) != tolower(s[i - t]))
			return 1;

	return 0;
}

int firestring_estr_strcasecmp(struct firestring_estr_t *f, const char *s) {
	int i;

	if (f->l != strlen(s))
		return 1;

	for (i = 0; i < f->l; i++)
		if (tolower(f->s[i]) != tolower(s[i]))
			return 1;

	return 0;
}

int firestring_estr_strcmp(struct firestring_estr_t *f, const char *s) {
	int i;

	if (f->l != strlen(s))
		return 1;

	for (i = 0; i < f->l; i++)
		if (f->s[i] !=s[i])
			return 1;

	return 0;
}

int firestring_estr_strcpy(struct firestring_estr_t *f, const char *s) {
	int l;

	l = strlen(s);
	
	if (f->a < l)
		return 1;

	f->l = l;
	memcpy(f->s,s,l);
	return 0;
}

int firestring_estr_strcat(struct firestring_estr_t *f, const char *s) {
	int l;

	l = strlen(s);
	
	if (l + f->l > f->a)
		return 1;

	memcpy(&f->s[f->l],s,l);
	f->l += l;
	return 0;
}

int firestring_estr_estrcasecmp(struct firestring_estr_t *t, struct firestring_estr_t *f, int start) {
	int i;

	if (f->l != t->l - start)
		return 1;
	
	for (i = 0; i < f->l; i++)
		if (tolower(f->s[i]) != tolower(t->s[i + start]))
			return 1;

	return 0;
}

int firestring_estr_estrncasecmp(struct firestring_estr_t *t, struct firestring_estr_t *f, int length, int start) {
	int i;

	for (i = 0; i < length; i++)
		if (tolower(f->s[i]) != tolower(t->s[i + start]))
			return 1;

	return 0;
}

int firestring_estr_estrcpy(struct firestring_estr_t *t, struct firestring_estr_t *f, int start) {
	if (f->l - start > t->a)
		return 1;

	t->l = f->l - start;
	memcpy(t->s,&f->s[start],t->l);

	return 0;
}

int firestring_estr_estrcmp(struct firestring_estr_t *t, struct firestring_estr_t *f, int start) {
	int i;

	if (f->l != t->l - start)
		return 1;

	for (i = 0; i < f->l; i++)
		if (f->s[i] != t->s[i + start])
			return 1;

	return 0;
}

int firestring_estr_estrcat(struct firestring_estr_t *t, struct firestring_estr_t *f, int start) {
	if (f->l - start + t->l > t->a)
		return 1;

	memcpy(&t->s[t->l],&f->s[start],f->l - start);
	t->l += f->l - start;

	return 0;
}

int firestring_estr_estrstr(struct firestring_estr_t *haystack, struct firestring_estr_t *needle, int start) {
	int i,l;

	l = haystack->l - needle->l;
	for (i = start; i < l; i++) {
		if (memcmp(&haystack->s[i],needle->s,needle->l) == 0)
				return i;
	}

	return -1;
}

int firestring_estr_estristr(struct firestring_estr_t *haystack, struct firestring_estr_t *needle, int start) {
	int i;
	int j;
	
	for (i = start; i <= haystack->l - needle->l; i++) {
		for (j = 0; j < needle->l; j++)
			if (tolower(haystack->s[i+j]) != tolower(needle->s[j]))
					break;
		if (j == needle->l)
			return i;
	}

	return -1;
}

struct firestring_conf_t *firestring_conf_parse(const char *filename) {
	struct firestring_conf_t *list = NULL;
	FILE *fp;
	char buf[512];
	char *var, *val, *end;
	char *prev_var = NULL, *prev_val = NULL;

	fp = fopen(filename, "r");
	if (!fp)
		return NULL;

	while (fgets(buf, sizeof (buf), fp) != NULL) {
		/* break into variable and value */
		if (prev_var == NULL) {
			var = firestring_chug(buf);
			if (*var == '#')
				continue;	/* comment */

			val = strchr(var,'=');
			if (val == NULL)
				continue;	/* malformed line */
			*val++ = '\0';

			val = firestring_chug(val);
			var = firestring_chomp(var);
		} else {
			/* continuing multi-line */
			var = prev_var;
			val = firestring_concat(prev_val, buf, NULL);
			free(prev_val);
			prev_val = NULL;
		}

		/* handle quoted and multi-line values */
		if (*val == '"') {
			end = strrchr(val, '"');
			if (end == val || *(end - 1) == '\\') {
				val = try_escaped_newline(val);
				prev_var = firestring_strdup(var);
				prev_val = firestring_strdup(val);
				continue;
			}
			val++;
			*end = '\0';
		}
		else
			val = firestring_chomp(val);

		/* save variable/value to list */
		list = firestring_conf_add(list, var, val);
		if (prev_var != NULL) {
			free(prev_var);
			prev_var = NULL;
		}
	}

	fclose (fp);
	return list;
}

struct firestring_conf_t *firestring_conf_add(struct firestring_conf_t *next, const char *var, const char *value) {
	struct firestring_conf_t *conf;

	if (var != NULL && value != NULL) {
		conf = firestring_malloc(sizeof(*conf));
		conf->next = next;
		conf->var = firestring_strdup(var);
		conf->value = firestring_strdup(value);
	} else
		conf = next;

	return conf;
}

char *firestring_conf_find(struct firestring_conf_t *config, const char *var) {
	while (config != NULL) {
		if (firestring_strcasecmp(config->var, var) == 0)
			return config->value;
		config = config->next;
	}

	return NULL;
}

char *firestring_conf_find_next(struct firestring_conf_t *config, const char *var, const char *prev) {
	int canreturn = 0;

	if (prev == NULL)
		canreturn = 1;

	while (config != NULL) {
		if (firestring_strcasecmp(config->var, var) == 0) {
			if (canreturn == 1)
				return config->value;
			else if (config->value == prev)
				canreturn = 1;
		}
		config = config->next;
	}

	return NULL;
}

void firestring_conf_free(struct firestring_conf_t *config) {
	struct firestring_conf_t *next;

	while (config != NULL) {
		next = config->next;
		free(config->var);
		free(config->value);
		free(config);
		config = next;
	}
}
